/*
 * Copyright (c) 1997, 1998, 2000, 2004 University of Michigan, Ann Arbor.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of Michigan, Ann Arbor. The name of the University 
 * may not be used to endorse or promote products derived from this 
 * software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Author: Sugih Jamin (jamin@eecs.umich.edu)
*/
#ifndef lint
static char rcsid[] =
"@(#) $Id: _vip.c,v 1.10 2000/10/13 21:18:44 sugih Exp $ (UM)";
#endif

#include <stdio.h>
#include <sys/errno.h>
#include <sys/types.h>

#include "vnet.h"
#include "vip.h"
#include "vcmp.h"
#include "vrp.h"
#include "vtp.h"

extern int errno;

/*
 * vip_input(vif, vrt)
 *
 * A VIP pkt arrived on the virtual interface vif.
 * If the pkt carries a bad version number,
 * clears the pipe of any remaining bits.
 * This could remove some good pkts, too bad;
 * this is, after all, a connectionless network.
 *
 * If there is any error in receiving the pkt,
 * throw away whatever we have so far and return.
 *
 * If pkt is bona fide, check whether we are the
 * destination.  If so, figures out its next level
 * protocol and pass it up.
 *
 * If pkt is not for us, pack it up in a vmbuf_t and
 * send it along to its destination by calling vip_forw,
 * passing along the virtual router vrt.
 * 
 * Doesn't return anything.
*/
void
_vip_input(vif_t *vif, vrouter_t *vrt)
{
  int err;
  int plen;
  vip_t vip;
  char *buf;
  vmbuf_t vmbuf;

  /* pull up vip header */
  err = vif_pullupn(vif, (char *) &vip, sizeof(vip_t));
  if (err < 0 && (errno == EPIPE || errno == EWOULDBLOCK)) {
    /* connection closed by seeder or full pkt never arrives
       throw pkt away and return. */
    return;
  }
  vnet_assert(err < 0, "vip_input: bad pkt");

  /* if unknown version */
  if (ntohs(vip.vip_ver) != VIP_VERSION) {
    
    /* clear interface of bits (see comments below) */
    buf = (char *) malloc(VNET_BLEN);
    do {
      err = vif_pullup(vif, buf, VNET_BLEN);
    } while (!(err < 0 && (errno == EWOULDBLOCK || errno == EPIPE)));
    free(buf);
    
    if (err > 0) {
      /* dump vip to screen to help debugging */
      vip_drop(vrt, &vip, 0, VCMP_ELAST);
    }
    
    return;
  }

  /* pull up the rest of the packet */
  plen = ntohs(vip.vip_len);
  buf = (char *) malloc(plen);
  err = vif_pullupn(vif, buf, plen);
  if (err < 0 && (errno == EPIPE || errno == EWOULDBLOCK)) {
    /* vif down by seeder or full pkt never arrives.
       Throw pkt away and return. */
    free(buf);
    return;
  }
  vnet_assert(err < 0, "vip_input: data pullup");

  if (vrp_isself(vrt, &vip.vip_dst)) {
    /* if pkt addressed to self, strip off vip header and either: 
       - give data portion to vcmp_input() if pkt->vp_proto is VIP_PROTOVCMP,
       - give data portion to vtp_input() if pkt->vp_proto is VIP_PROTOVTP,
       - give data portion to vrp_update() if pkt->vp_proto is VIP_PROTOVRP.*/

    vip.vip_prot = ntohs(vip.vip_prot);
    switch(vip.vip_prot) {
    case VIP_PROTOVRP:
      vrp_update(vrt, vif, (vrp_t *) buf, plen);
      break;
    case VIP_PROTOVTP:
      vtp_input(&vip.vip_src, buf, plen);
      break;
    case VIP_PROTOVCMP:
      vcmp_input(&vip.vip_src, (vcmp_t *) buf);
      break;
    default:
      vip_drop(vrt, &vip, 0, VCMP_EPROT);
      break;
    }
  } else {
    vmbuf.vmbuf_data = buf;
    vmbuf.vmbuf_len = plen;
    vmbuf.vmbuf_next = 0;

    /* forward packet. */
    vip_forw(vrt, &vip, &vmbuf);
  }
  free(buf);

  return;

  /*
   * How to clear interface of bits in case of error:
   *   Use vif_pullup() until the pipe is empty.
   *
   * > The big question is how much do I pull up and how do I make
   * > sure that I don't start pulling up the header for the next packet.
   *
   * You can't be sure that you're not pulling up the header for the
   * next packet.  What we are doing here is called "Panic-mode recovery" 
   * in the parlance of parser construction.  Here's the definition
   * from the dragon book (Aho, Sethi, and Ullman, _Compilers: Principles,
   * Techniques, and Tools_, p. 164), "On discovering an error, the parser 
   * discards input symbols one at a time until one of a designated set of
   * synchronizing tokens is found.  The synchronizing tokens are usually
   * delimiters, such as semicolon or *end*, whose role in the source
   * program is clear."  In our code, we use an empty pipe as the
   * synchronizing "token"/delimiter.  Put another way, consider a
   * push-down automaton, we recognize an empty pipe as the symbol 
   * denoting end of input that allows us to accept the input and pop 
   * the stack.
   *
   * Empty pipe is detected when vif_pullup() returns -1 with errno set to
   * EWOULDBLOCK or EPIPE.
   */
}
  
/*
 * vip_forw(vrt, vp, vmbp)
 *
 * vrt is the virtual router.
 * vp is the original VIP header of this pkt,
 * in network byte order.
 * vmbp is the vmbuf chain of the remainder of
 * the pkt.
 *
 * Decrements the pkt's hop limit.
 * If hop limit exceeded, drop pkt,
 * else, routes pkt.
 * If route found, forwards pkt
 * to next hop, else drop pkt.
 * Send off VCMP error pkts where appropriate.
*/
void
_vip_forw(vrouter_t *vrt, vip_t *vp, vmbuf_t *vmbp)
{
  int err;
  vif_t *outf;
  vmbuf_t vmbuf;
  
  vnet_assert(!vmbp, "vip_forw: empty pkt");
  
  /* decrement hop limit */
  vp->vip_hop = ntohs(vp->vip_hop)-1;

  if (vp->vip_hop > 0) {
    vp->vip_hop = htons(vp->vip_hop);

    /* route pkt */
    outf = vrp_route(vrt, &vp->vip_dst, 0);
    if (outf) {
      vmbuf.vmbuf_data = (char *) vp;
      vmbuf.vmbuf_len = sizeof(vip_t);
      vmbuf.vmbuf_next = vmbp;

      err = vif_output(outf, VIF_PROTOVIP, &vmbuf);
      if (err < 0) {
        /* drop pkt on transmission error */
        vip_drop(vrt, vp, vmbp->vmbuf_data, VCMP_ELINK);
      } 
    } else {
      /* drop pkt destination unreachable */
      vip_drop(vrt, vp, vmbp->vmbuf_data, VCMP_EHOST);
    }
  } else {
    /* drop pkt hop limit exceeded */
    vip_drop(vrt, vp, vmbp->vmbuf_data, VCMP_EHOPS);
  }

  return;
}
    
/*
 * vip_output(vrt, dst, ttl, proto, vmbp)
 *
 * Creates a VIP header with the args.
 * Sends it out to dst along with the vmbuf chain vmbp
 * by calling vif_output.  If ttl is 0, figure out
 * the actual number of hop to dst, otherwise, use the   
 * given ttl as the hop limit.
 *
 * Return 0 if no error, otherwise -1 and errno set to:
 * EMSGSIZE: pkt too large,
 * EHOSTUNREACH: dst unreachable,
 * or whatever vif_output sets it to.
*/
int
_vip_output(vrouter_t *vrt, vin_addr_t *dst, u_short ttl, u_short proto, vmbuf_t *vmbp)
{
  vip_t vip;
  vmbuf_t vmbuf, *pkt;
  int plen, err;
  vif_t *outf;

  vnet_assert(!vmbp, "vip_output: empty pkt");

  plen = 0;
  for (pkt = vmbp; pkt; pkt = pkt->vmbuf_next) {
    vnet_assert(!pkt->vmbuf_data, "vip_output: vmbuf data points to NULL");
    plen += pkt->vmbuf_len;
  }
  if (plen > VIP_MAXPKTSIZE) {
    errno = EMSGSIZE;
    return(-1);
  }
  
  /* find route to destination */
  outf = vrp_route(vrt, dst, &vip.vip_hop);
  if (!outf) {
    errno = EHOSTUNREACH;
    return(-1);
  }

  /* Make a vip packet: create the header and append data after it.
     This is called marshalling/serialization. */
  vip.vip_ver = htons(VIP_VERSION);
  vip.vip_hop = ttl ? htons(ttl) : htons(vip.vip_hop);
  vip.vip_len = htons(plen);
  vip.vip_prot = htons(proto);
  memcpy((char *) &vip.vip_src, (char *) vrp_self(vrt), sizeof(vin_addr_t));
  memcpy((char *) &vip.vip_dst, (char *) dst, sizeof(vin_addr_t));

  vmbuf.vmbuf_data = (char *) &vip;
  vmbuf.vmbuf_len = sizeof(vip_t);
  vmbuf.vmbuf_next = vmbp;

  /* send packet out */
  err = vif_output(outf, VIF_PROTOVIP, &vmbuf);
  if (err >= 0) {
    vnet_assert(err < (plen + sizeof(vip_t)), 
                "vip_output: vif_output sent truncated packet");
    err = 0;
  } 

  return(err);
}

/*
 * vip_drop(vrt, vp, data, errcode)
 *
 * Prints to screen the content of the VIP header pointed to by vp.
 * If data is not null, pass it on to the higher level protocol
 * it belongs to.  Generates a VCMP error message where appropriate,
 * with errcode as the VCMP error code defined in vcmp.h.
 * If errcode is VCMP_ELAST, don't generate a VCMP error message.
*/ 
void
_vip_drop(vrouter_t *vrt, vip_t *vp, char *data, u_short errcode)
{
  int err, proto;
  vmbuf_t vmbuf;
  
  fprintf(stderr, "VIP drop: vcmp errmsg: %s\n", vcmp_errmsg(errcode));

  fprintf(stderr, "ver: %d; hlim: %d\n", ntohs(vp->vip_ver), ntohs(vp->vip_hop));
  fprintf(stderr, "plen: %d; proto: %d\n", ntohs(vp->vip_len), ntohs(vp->vip_prot));
  fprintf(stderr, "saddr: %s.%d; ",
          inet_ntoa(vp->vip_src.vad_ipv4), ntohs(vp->vip_src.vad_port));
  fprintf(stderr, "daddr: %s.%d\n\n",
          inet_ntoa(vp->vip_dst.vad_ipv4), ntohs(vp->vip_dst.vad_port));

  vmbuf.vmbuf_data = (char *) vp;
  vmbuf.vmbuf_len = sizeof(vip_t);
  vmbuf.vmbuf_next = 0;
  
  err = 0;
  if (data) {
    proto = ntohs(vp->vip_prot);
    switch (proto) {
    case VIP_PROTOVCMP:
      fprintf(stderr, "Dropping VCMP pkt\n");
      vcmp_input(&vp->vip_src, (vcmp_t *) data);
      break;
    case VIP_PROTOVRP:
      vrp_drop((vrp_t *) data, ntohs(vp->vip_len));
      err = vcmp_output(vrt, &vp->vip_src, errcode, &vmbuf);
      break;
    case VIP_PROTOVTP:
      vtp_drop(&vp->vip_src, data, ntohs(vp->vip_len));
      err = vcmp_output(vrt, &vp->vip_src, errcode, &vmbuf);
      break;
    default:
      fprintf(stderr, "Unknown protocol type %d\n", ntohs(vp->vip_prot));
      err = vcmp_output(vrt, &vp->vip_src, errcode, &vmbuf);
      break;
    }
  } else if (errcode != VCMP_ELAST) {
    err = vcmp_output(vrt, &vp->vip_src, errcode, &vmbuf);
  }

  if (err) {
    /* right now doesn't do anything with vcmp_output error. */
  }
  
  return;
}
