/* Copyright (c) 1997, 1998, 2000, 2001, 2004 University of Michigan, Ann Arbor.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of Michigan, Ann Arbor. The name of the University 
 * may not be used to endorse or promote products derived from this 
 * software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Author: Sugih Jamin (jamin@eecs.umich.edu)
 *         Jatin Nahar (jnahar@eecs.umich.edu)
 *         Wenjie Wang (wenjiew@eecs.umich.edu)
*/
#ifndef lint
static char rcsid[] =
"@(#) $Id: seeder.c,v 1.15 2001/10/14 19:09:55 sugih Exp $ (UM)";
#endif

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <signal.h>
#include <fcntl.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include "vnet.h"
#include "vif.h"
#include "vip.h"
#include "vrp.h"
#include "vtp.h"

#define SDR_RAND 17         /* random number seed */
#define SDR_PROBUP   1      /* probability for bringing link up */
#define SDR_PROB     3      /* probability for doing some other stuff */
#define SDR_INCDEPTH 2      /* max includes depth in script file */

#define SDR_MINRTR 3        /* minimum number of router for testing vrp */

#define SDR_INFINITY VRP_INFINITY+1  /* for purposes of modifying the virtual */
#define SDR_DELETED VRP_INFINITY+2   /* network's topology */

#define SDR_MAXLINK 10        /* maximum number of link a node may have */
#define SDR_FWLINK 0         /* the direction of the connection */
#define SDR_BWLINK 1

extern int errno;

typedef struct _link link_t;
typedef struct _node node_t;

struct _node {
  vif_t nd_vif;                 /* the vif used to communicate with a vrouted */
  int nd_nlink;                 /* number of links the node has */
  link_t *nd_links[SDR_MAXLINK];/* the connectivity of the node */
};

struct _link {
  node_t *ln_nodev;             /* the two nodes that connect to a link */
  node_t *ln_nodew;
  u_short ln_cost;              /* the cost of the link, mostly 1 */
};

static link_t *freelinks;

void
sdr_usage(char *progname)
{
  fprintf(stderr, 
          "Usage: %s -n <number of routers> [ -p <portnum> -r <random seed> -f <file name> ]\n", progname); 
  exit(errno);
}

/*
 * sdr_args: Parse args.
 *
 * Return 0 on success or errno on failure.
 * On successful return, "port" contains the
 * port# to use, in host byte order; "nrouter"
 * contains the number of routers to expect;
 * "random" contains the random seed, "cmd" contains
 * 'f' if script file is specified, 'b' if batch
 * file is specified, '\0' otherwise.  "cmd" is
 * expected to be '\0' upon entry to the function.
 * "fname" contains the name of either the script
 * or batch file.
 * Nothing else is modified.
*/
int
sdr_args(int argc, char *argv[], int *port, int *nrouter, long *random, 
    char *fname)
{
  char c;
  extern int optind;
  extern char *optarg;

  errno = 0;
  *port = SDR_PORT;
  *nrouter = 0;
  *random = (long) SDR_RAND;
  fname[0] = '\0';

  while ((c = getopt(argc, argv, "p:n:r:f:")) != EOF) {
    switch (c) {
    case 'p':
      *port = atoi(optarg);
      break;
    case 'n':
      *nrouter = atoi(optarg);
      vnet_assert(*nrouter >= VRT_INITSIZE, "sdr_args: Too many vrouters");
      break;
    case 'r':
      *random = atol(optarg);
      break;
    case 'f':
      strcpy(fname, optarg);
      break;
    default:
      errno = EINVAL;
      break;
    }
  }

  return(errno);
}

/*
 * sdr_setup_conn: setup a TCP connection
 * bind it to port "port", listen on it.
 * qlen is the qlen to associate with the listening socket.
 *
 * Terminate process on error.
 * Else return the connected socket id.
*/
int
sdr_setup_conn(int port, int qlen)
{
  int sd, reuse, err;
  struct sockaddr_in self;
  
  /* create a TCP socket */
  sd = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
  vnet_assert(sd < 0, "sdr_setup_conn: socket");

  /* initialize socket address */
  memset((char *) &self, 0, sizeof(struct sockaddr_in));
  self.sin_family = AF_INET;
  self.sin_addr.s_addr = INADDR_ANY;
  self.sin_port = htons((u_short) port);

  /* reuse local address so that bind doesn't complain
     of address already in use. */
  reuse = 1;
  err = setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, (char *) &reuse, sizeof(int));
  vnet_assert(err, "sdr_setup_conn: setsockopt");
  
  /* bind address to socket */
  err = bind(sd, (struct sockaddr *) &self, sizeof(struct sockaddr_in));
  vnet_assert(err, "sdr_setup_conn: bind");

  /* listen on socket */
  err = listen(sd, qlen);
  vnet_assert(err, "sdr_setup_conn: listen");

  /* return socket id. */
  return(sd);
}

/*
 * sdr_waitfor_conn: wait for connection on the given socket, sd.
 *
 * The descriptor of the connected socket and the address of the
 * remote vrouted is stored in the provided vif.
 * Terminate process on error.
*/
int
sdr_waitfor_conn(int sd, vif_t *vif)
{
  int td, len;
  struct hostent *chost;
  struct sockaddr_in client;
  
  /* wait for and accept the new connection. */
  len = sizeof(struct sockaddr_in);
  td = accept(sd, (struct sockaddr *) &client, &len);
  vnet_assert(td < 0, "sdr_waitfor_conn: accept");

  /* get the host entry info on the connected host. */
  chost = gethostbyaddr((char *) &client.sin_addr, sizeof(struct in_addr), 
                        AF_INET);
  vnet_assert(!chost, "sdr_waitfor_conn: gethostbyaddr");
  /* for debug purposes */
  fprintf(stderr, "Connected from %s.%d (%s.%d)\n", 
          chost->h_name, ntohs(client.sin_port),
          inet_ntoa(client.sin_addr), ntohs(client.sin_port));

  vif->vif_outd = td;
  vif->vif_status = VIF_ONLINE;
  memcpy((char *) &vif->vif_vad.vad_ipv4, (char *) &client.sin_addr, 
         sizeof(struct in_addr));
  vif->vif_vad.vad_port = client.sin_port; /* in network byte order */
  vif->vif_vad.vad_metric = VIF_DIRECT;    /* in host byte order */

  return(td);
}

/*
 * sdr_dump():
 *
 * This function was written to aid in debugging.  
 * It prints out the network topology.
 * It is called at the initial topology setup and 
 * also after each randomization of the network.
 *
 * nrouter is the number of routers in the network
 * nodes is the pointer to the network nodes and links.
 * 
 * none of the data is altered, and there are no side effects.
 *
 * Authors: David Gray and Chris David (EECS 489, Winter 1997),
 * with contributions from Tim Maun and Mike Dunn 
 * (EECS 489, Fall 2001).
*/
void
sdr_dump(int nrouter, node_t *nodes)
{
  int i, j, k;
  vin_addr_t *vadp;
  struct hostent *host;

  fprintf(stdout, "\nN e t w o r k   T o p o l o g y\n");

  if (!nodes || !nrouter) {
    /* invalid data */
    fprintf(stderr, "sdr_dump: Invalid parameters\n");
  }

  for (i = 0; i < nrouter; i++) {
    /* for each virtual router, print out
     *  router ID number,
     *  host name on which it runs,
     *  IP of host and port number,
     *  number of links, and
     *  identity of links.
     */
    vadp = &nodes[i].nd_vif.vif_vad;
    host = gethostbyaddr((char*) &vadp->vad_ipv4, sizeof(struct in_addr), AF_INET);

    fprintf(stdout, "R%d: %s(%s).%d\tlinks = %d\n", i, host->h_name,
            inet_ntoa(vadp->vad_ipv4), ntohs(vadp->vad_port), nodes[i].nd_nlink);

    k = 0;
    for(j = 0; k < nodes[i].nd_nlink; j++) {
      /* print out each link */

      if (nodes[i].nd_links[j]) {

        fprintf(stdout, "\tL%d: ", j);
        if (vin_eq(&nodes[i].nd_vif.vif_vad,
                   &nodes[i].nd_links[j]->ln_nodev->nd_vif.vif_vad)) {
          vadp = &nodes[i].nd_links[j]->ln_nodew->nd_vif.vif_vad;
          host = gethostbyaddr((char*) &vadp->vad_ipv4, sizeof(struct in_addr), AF_INET);
          fprintf(stdout, "%s(%s).%d", host->h_name, inet_ntoa(vadp->vad_ipv4), ntohs(vadp->vad_port));
        } else {
          vadp = &nodes[i].nd_links[j]->ln_nodev->nd_vif.vif_vad;
          host = gethostbyaddr((char*) &vadp->vad_ipv4, sizeof(struct in_addr), AF_INET);
          fprintf(stdout, "%s(%s).%d", host->h_name, inet_ntoa(vadp->vad_ipv4), ntohs(vadp->vad_port));
        }
        fprintf(stdout, "\t%s\n",
                nodes[i].nd_links[j]->ln_cost == VIF_DIRECT ?
                "up" : "down");
        k++;
      }
    }
  }
  fflush(stdout);
}

/* sdr_gbclink: garbage collect old links.
 *
 * Down links are garbage collected.
*/
void
sdr_gbclink(int nrouter, node_t *nodes)
{
  int i,j ;
  link_t *lp;

  for (i = 0; i < nrouter; i++) {
      
    for (j = 0; j < SDR_MAXLINK; j++) {
      if (lp = nodes[i].nd_links[j]) {
         
        if (lp->ln_cost == VIF_LINKDOWN) {
          /* SDR_INFINITY means link is down,
             VIF_LINKDOWN means link down has 
               been communicated to the vrouters.
             SDR_DELETED means one end of the link
               has cleaned itself up. */
          lp->ln_cost = SDR_DELETED;
          nodes[i].nd_links[j] = 0;
          nodes[i].nd_nlink--;
        } else if (lp->ln_cost == SDR_DELETED) {
          /* if marked deleted by the adjacent node, remove link */
          lp->ln_nodev = (node_t *) freelinks;
          freelinks = lp;
          nodes[i].nd_links[j] = 0;
          nodes[i].nd_nlink--;
        }
      }
    }
  }

  return;
}

/*
 * sdr_newlink: returns pointer to a new link_t.
 *
 * Old/unused links from nodes are collected
 * and reused.
 */
link_t *
sdr_newlink(int nrouter, node_t *nodes) 
{
  link_t *lp;
 
  if (!freelinks) {
    sdr_gbclink(nrouter,nodes);
  }

  if (!freelinks) {
    /* memory allocation */
    lp = (link_t *) malloc(sizeof(link_t));
    vnet_assert(!lp, "sdr_newlink: Out of memory");
  } else {
    lp = freelinks;
    freelinks = (link_t *) lp->ln_nodev;
  }    

  return lp;
}
 
/*
 * sdr_top: create a simple linked-list topology
 * of the nrouter pointed to by nodes.
*/
void
sdr_top(int nrouter, node_t *nodes)
{
  int i;
  link_t *links;

  links = (link_t *) malloc((nrouter-1)*sizeof(link_t));

  for (i = 0; i < nrouter-1; i++) {
    links[i].ln_nodev = &nodes[i];
    links[i].ln_nodew = &nodes[i+1];
    links[i].ln_cost = VIF_DIRECT;

    nodes[i].nd_links[SDR_FWLINK] = &links[i];
    nodes[i].nd_nlink++;
    nodes[i+1].nd_links[SDR_BWLINK] = &links[i];
    nodes[i+1].nd_nlink++;
  }

  return;
}

/*
 * sdr_adj: return the node adjecent to "nd"
 * on link "lp."
 */
node_t *
sdr_adj(node_t *nd, link_t *lp)
{
  node_t *np;

  np = nd;
  if (lp->ln_nodev != np) {
    np = lp->ln_nodev;
  } else if (lp->ln_nodew != np) {
    np = lp->ln_nodew;
  }

  return (np);
}
  
/*
 * sdr_isadj(v,w,nodes)
 *
 * Returns a pointer to the link connecting v and w
 * if v and w are neighbors, 0 otherwise.  No side effect.
*/
link_t *
sdr_isadj(node_t *v, node_t *w)
{
  int i;
  
  if (w->nd_nlink < SDR_MAXLINK) {
    for (i = 0; i < SDR_MAXLINK; i++) {
      if (v->nd_links[i] && 
          (v->nd_links[i]->ln_nodev == w || 
           v->nd_links[i]->ln_nodew == w)) {
        return (v->nd_links[i]);
      }
    }
  }

  return (0);
}

/*
 * sdr_rand: randomized the topology of the
 * virtual network.
*/
void
sdr_rand(int nrouter, node_t *nodes)
{
  int i, j, k;
  int adj;
  link_t *lp;
  node_t *np;

  /* first garbage collect all down links */
  sdr_gbclink(nrouter, nodes);

  for (i = 0; i < nrouter; i++) {

    np = &nodes[i];
    
    for (j = 0; j < SDR_MAXLINK; j++) {

      /* if neighbor */
      if (lp = np->nd_links[j]) {

        /* probabilistically mark link down. */
        if ((lp->ln_cost == VIF_DIRECT) 
            && (lrand48()%10 < SDR_PROB)) { 
          lp->ln_cost = SDR_INFINITY;
        }

      } else if ((np->nd_nlink < (nrouter-1)) && (lrand48()%10 < SDR_PROBUP)) {
        /* if not fully connected, probabilistically create a new connection */

        do {
          do { /* find a target node that is not self */
            adj = lrand48()%nrouter;
          } while (adj == i);

          /* loop till a target node that is not already a neighbor is found */
        } while (sdr_isadj(np, &nodes[adj]));
        
        /* find an empty link slot */
        for (k = 0; k < SDR_MAXLINK && nodes[adj].nd_links[k]; k++);
        vnet_assert(k >= SDR_MAXLINK, "sdr_rand: bad adjacency");
        
        /* and link up self with target node */
        lp = sdr_newlink(nrouter, nodes);
        lp->ln_nodev = np;
        lp->ln_nodew = &nodes[adj];
        lp->ln_cost = VIF_LINKUP;
        np->nd_links[j] = lp;
        np->nd_nlink++;
        nodes[adj].nd_links[k] = lp;
        nodes[adj].nd_nlink++;
      }
    }
  }

  return;
}

/*
 * sdr_toggle(v, w, up)
 *
 * Add a link between v and w or bring it down.
 * If a new link is to be added but v or w already 
 * has SDR_MAXLINK, kill program.
 *
 * Return 1 if topology changed, 0 otherwise.
*/
int
sdr_toggle(node_t *v, node_t *w, int up, int nrouter, node_t *nodes)
{
  int i, j;
  link_t *lp;
  int changed;
  
  changed = 0;
  lp = sdr_isadj(v, w);
      
  if (!up) { /* toggle down */
    if (lp) { /* not already down */
      lp->ln_cost = SDR_INFINITY;
      changed = 1;
    }
  } else {

    if (!lp) {  /* link does not exist so create a new link */
    
      /* find an empty link slot for v and w */
      for (i = 0; i < SDR_MAXLINK && v->nd_links[i]; i++);
      vnet_assert(i >= SDR_MAXLINK,
                  "sdr_toggle: bad adjacency.  Node number must be between 0 and n-1");
      for (j = 0; j < SDR_MAXLINK && w->nd_links[j]; j++);
      vnet_assert(j >= SDR_MAXLINK,
                  "sdr_toggle: bad adjacency.  Node number must be between 0 and n-1");
      
      /* link up v and w */
      lp = sdr_newlink(nrouter, nodes);
      
      lp->ln_nodev = v;
      lp->ln_nodew = w;
      lp->ln_cost = VIF_LINKUP;
      v->nd_links[i] = lp;
      v->nd_nlink++;
      w->nd_links[j] = lp;
      w->nd_nlink++;
      
      changed = 1;
    } else if (lp->ln_cost != VIF_DIRECT) {
      lp->ln_cost = VIF_LINKUP;
      changed = 1;
    }
  }

  return (changed);
} 

/*
 * sdr_seed: for each node on the virtual network,
 * create a distance vector for each of the node's
 * links, marshall them up in a vrp packet, and
 * send it to the node.
 * nrouter is the number of nodes in the virtual network.
 * nodes points to the list of nodes.
 * self identifies the seeder, in network byte-order.
*/
void
sdr_seed(int nrouter, node_t *nodes)
{
  link_t *lp;
  node_t *np;
  vrc_t *seeds;
  int i, j, k, plen;
  vin_addr_t *dvp;
  vmbuf_t vmbuf;
  
  for (i = 0; i < nrouter; i++) {

    /* if node is connected */
    if (nodes[i].nd_nlink) {
      /* create the vrp packet */
      plen = sizeof(vrc_t) - sizeof(vin_addr_t) + 
             nodes[i].nd_nlink*sizeof(vin_addr_t);
      seeds = (vrc_t *) malloc(plen);
      seeds->vrc_vers = htons(VRC_VERSION);
      seeds->vrc_nvad = htons(nodes[i].nd_nlink);
      dvp = (vin_addr_t *) &seeds->vrc_vad;
      
      for (j = 0, k = 0; j < SDR_MAXLINK; j++) {
        /* marshall the distance vector of the links */
        if (lp = nodes[i].nd_links[j]) {
          np = sdr_adj(&nodes[i], lp);
          vnet_assert(nrouter > 1 && np == &nodes[i], 
                      "sdr_seed: bad adjacency");
          vnet_assert(k >= nodes[i].nd_nlink, 
                      "sdr_seed: inconsistent link count");
          memcpy((char *) &dvp[k], (char *) &np->nd_vif.vif_vad, 
                 sizeof(vin_addr_t));
          
          if (lp->ln_cost == SDR_INFINITY) {
            /* SDR_INFINITY means link is down,
               VIF_LINKDOWN means link down has been communicated
               to the vrouters. */
            lp->ln_cost = VIF_LINKDOWN;
          } else if (lp->ln_cost == VIF_LINKUP) {
            lp->ln_cost = VIF_DIRECT;
          }
          dvp[k].vad_metric = htons(lp->ln_cost);
          k++;
        }
      }
      /* send the vif pkt off to the node */
      vmbuf.vmbuf_data = (char *) seeds;
      vmbuf.vmbuf_len = plen;
      vmbuf.vmbuf_next = 0;
      vif_output(&nodes[i].nd_vif, VIF_PROTOVRC, &vmbuf);
      free(seeds);
    }
  }
  
  return;
}

/*
 * sdr_parsecript: determinitically create a topology of the virtual network.
 * Deterministic topology can be created through console input,
 * or by specifying a file containing those commands.
*/
int
sdr_parsescript(FILE *fp, int nrouter, node_t *nodes, int iter)
{
  int i, v, w, up;
  int changed;
  char buf[VNET_BLEN];
  FILE *ifp;

  vnet_assert(!fp, "sdr_parsescript: bad STREAM");
  vnet_assert(iter > SDR_INCDEPTH, "sdr_parsescript: includes too deep");
  
  changed = 0;
  
  while (fgets(buf, VNET_BLEN, fp)) {

    /*
     * End of changes is indicated by a period. 
     * Line starting with '!', denotes a warning that has to be printed out.
     * Line starting with '#' denotes either a line of comments,
     *   in which case the whole line is ignored, or an #include
     *   directive, in which case the #include "file" is parsed.
     * Line starting with 's' or ',' tells the seeder to send out
     *   the seeds and wait for the new topology to stabilize.
     *   The wait time (in seconds) may be specified.  If not
     *   specified, the wait time is a multiple of nrouter.
     */

    switch (buf[0]) {
    case '!':
      /* print out warning */
      fprintf(stdout, "%s", &buf[1]);
      break;
    case '#':
      /* comment line or #include */
      for (i = 1; buf[i] != ' ' && buf[i] != '\n'; i++);
      if (buf[i] != '\n') {
        buf[i] = '\0';
        if (!strcmp(&buf[1], "include")) {
          /* skip the '\0' and '"' */
          i += 2;
          /* skip the matching '"' and '\n' */
          buf[i+strlen(&buf[i])-2] = '\0';
          ifp = fopen(&buf[i], "r" );
          vnet_assert(!ifp, "sdr_parsescript: ");
          fprintf(stderr, "sdr_parsescript: parsing %s\n", &buf[i]);
          changed |= sdr_parsescript(ifp, nrouter, nodes, iter+1);
          fclose(ifp);
        }
      }
      break;
    case ',':
      /* send out seeds */
      if (changed) {
        sdr_seed(nrouter, nodes);
        sdr_dump(nrouter, nodes);
        changed = 0;
        
        /* wait some seconds or, if no second is given,
           wait until new topology stabilized */
        fprintf(stderr, "Reconfiguring to above network topology . . . ");
        for (i = 1; buf[i] != ' ' && buf[i] != '\n'; i++);
        if (buf[i] != '\n') {
          buf[strlen(buf)-1] = '\0';
          sleep(atoi(&buf[i+1]));
        } else {
          sleep((nrouter-1)*VRP_UPDTSECS);
        }
      }
      fprintf(stderr, "done.\nContinuing with script . . . .\n");
      break;
    case '.':
      fflush(fp);
      return (changed);
    case ' ':
    case '\t':
    case '\n':
      /* skip empty line */
      break;
    default:
      sscanf(buf, "%d %d %d", &v, &w, &up);
      if (v < 0 || v >= nrouter || w < 0 || w >= nrouter || v == w) {
        fprintf(stderr, "sdr_parsescript: node number out of bounds.\n");
      } else if (sdr_toggle(&nodes[v], &nodes[w], up, nrouter, nodes)) {
        changed |= 1;
      } else {
        fprintf(stderr, "Link (%d, %d) not changed.\n", v, w);
      }
      break;
    }
  }
  
  fflush(fp);
  return (changed);
}

int 
main(int argc, char *argv[])
{
  int port, nrouter;
  long random;
  char fname[VNET_BLEN];
  FILE *fp;

  int i, sd;
  node_t *nodes;
  vrouter_t vrt;

  char prev;         /* previous command */
  char buf[VNET_BLEN];

  int changed, wait;
  struct timeval ts1, ts2;


  if (sdr_args(argc, argv, &port, &nrouter, &random, fname)) {
    sdr_usage(argv[0]);
  }

  if (nrouter < 1 || nrouter > VRP_INFINITY-1) {
    errno = EINVAL;
    sdr_usage(argv[0]);
  }
  fprintf(stderr, "%s -n %d [ -p %d -r %ld -f %s]\n", 
          argv[0], nrouter, port, random, fname[0] ? fname : "<none>");
  srand48(random);           /* seeds the random number generator */

  signal(SIGPIPE, SIG_IGN);        /* don't die just because client is dead */

  vrp_init(&vrt, port, 0);
  
  /* setup and listen on connection */
  sd = sdr_setup_conn(port, nrouter);

  /* allocate memory for nodes */
  nodes=(node_t *) calloc(nrouter, sizeof(node_t));

  for (i = 0; i < nrouter; i++) {
    /* wait for each vrouted to register */
    sdr_waitfor_conn(sd, &nodes[i].nd_vif);
    
    sleep(2);    /* wait for the vrouted to initialize its vifs */

    /* make a fake vrouter with routing table for seeder so that
       vnet routines can be used as is. */
    memcpy(&vrt.vrouter_rt[i+1].vrtbl_dst, &nodes[i].nd_vif.vif_vad, sizeof(vin_addr_t));
    vrt.vrouter_rt[i+1].vrtbl_vif = &nodes[i].nd_vif;
    
    /* test vtp_output down and up to vtp_input */
    sprintf(buf, "vrouted@%s.%d registered.\n",
            inet_ntoa(nodes[i].nd_vif.vif_vad.vad_ipv4),
            ntohs(nodes[i].nd_vif.vif_vad.vad_port));

    vtp_output(&vrt, &nodes[i].nd_vif.vif_vad, 0, buf, strlen(buf)+1);
  }

  if (nrouter > 1) {

    if (!fname[0]) {
      /* If no script was given, create a simple chained virtual network. */
      sdr_top(nrouter, nodes);
    } else {
      fp = fopen(fname, "r");
      vnet_assert(!fp, argv[0]);
      sdr_parsescript(fp, nrouter, nodes, 0);
    }

    changed = 1;
    prev = 'r';

    do {

      /* Propagate only when there is an actual change to the topology. */
      if (changed) { 
        /* send out seeds. Test vrp_update, vrp_propagate, and vip_output. */
        gettimeofday(&ts1, NULL);
        sdr_seed(nrouter, nodes);
        sdr_dump(nrouter, nodes);   /* print out current topology */
        changed = 0;
      }

      fprintf(stderr,
              "\nContinue testing (random, scriptfile, manual, quit) [%c]? ",
              prev);
      if (!(fgets(buf, VNET_BLEN, stdin)) || 
          (buf[0] == 'n') || (buf[0] == 'N') || 
          (buf[0] == 'q') || (buf[0] == 'Q')) {
        break;
      }
      fflush(stdin);
      
      /* make previous command the default */
      if (buf [0] == ' ' || buf[0] == '\n') {
        buf[0] = prev;
      } else {
        prev = buf[0];
      }

      switch(buf[0]) {
      case 'r':
      case 'R':
        /* randomize topology */
        sdr_rand(nrouter, nodes);
        changed = 1;
        break;
      case 'm':
      case 'M':
        /* reconfigure the network manually */
        fprintf(stderr, "Enter the node numbers and 1 or 0 (up or down),\n" );
        fprintf(stderr, "for example, 1 2 0 means bring link (1,2) down.\n" );
        fprintf(stderr, "End with '.' on a line by itself.\n\n");
        changed = sdr_parsescript(stdin, nrouter, nodes, 0);
        break;
      case 's':
      case 'S':
        fprintf(stderr, "\nEnter file name: ");
        fgets(fname, VNET_BLEN, stdin);
        fname[strlen(fname)-1] = '\0';
        fp = fopen(fname, "r" );
        if (!fp) {
          perror(fname);
          continue;
        }
        changed = sdr_parsescript(fp, nrouter, nodes, 0);
        fclose(fp);
        break;
      default:
        fprintf(stderr, "%c: Unknown command.\n", buf[0]);
        changed = 0;
        break;
      }

      if (changed) {
        fprintf(stderr, "Reconfiguring network topology . . . .\n");
        /* wait for network to stabilize in between 2 sets of updates*/
        gettimeofday(&ts2, NULL);
        wait = ((nrouter-1)*VRP_UPDTSECS) - time_diff(ts1, ts2);
        if (wait > 0) {
          sleep(wait);
        }
      }
      
    } while (1);
  }

  fprintf(stderr, "Done testing.  Bye!\n");
  
  exit(0);
}
