/*
 * Copyright (c) 1997, 1998, 2000, 2001, 2004 University of Michigan, Ann Arbor.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of Michigan, Ann Arbor. The name of the University 
 * may not be used to endorse or promote products derived from this 
 * software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Author: Sugih Jamin (jamin@eecs.umich.edu)
*/
#ifndef lint
static char rcsid[] =
"@(#) $Id: vif.c,v 1.18 2001/09/30 20:14:01 sugih Exp $ (UM)";
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <netdb.h>

#include "vnet.h"
#include "vif.h"
#include "vip.h"

#define VIF_PULLUPUSECS 500000
#define VIF_PULLUPMAXTRIES 10

extern int errno;

static int vif_input_nready;          /* number of ready sockets */
static fd_set vif_input_rset;         /* select read set */

/*
 * vif_init(rcd, port, vit):
 *
 * Initializes virtual interface table vit, makes this vrouter 
 * listen for connection from other vrouters.
 *
 * rcd is the socket descriptor connected to the seeder.  port is the
 * port number this vrouter uses to listen for connection from other
 * vrouters.
 *
 * The input descriptor of the first vif
 * connects to the seeder.  The output descriptor of the first vif
 * awaits connection from other vrouters.
 *
 * Terminates process on error.  Else returns nothing.
*/
void
vif_init(int rcd, int port, vit_t *vit)
{
  int sd;
  int flags, reuse, err;
  struct sockaddr_in self;
  
  /* don't die just because the other side is dead */
  signal(SIGPIPE, SIG_IGN);        /* don't die just because client is dead */

  /* create a TCP socket */
  sd = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
  vnet_assert(sd < 0, "vif_init: socket");
 
  /* initialize socket address */
  memset((char *) &self, 0, sizeof(struct sockaddr_in));
  self.sin_family = AF_INET;
  self.sin_addr.s_addr = INADDR_ANY;
  self.sin_port = htons((u_short) port);
 
  /* reuse local address so that bind doesn't complain
     of address already in use. */
  reuse = 1;
  err = setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, (char *) &reuse, sizeof(int));
  vnet_assert(err, "vif_init: setsockopt");
  
  /* bind address to socket */
  err = bind(sd, (struct sockaddr *) &self, sizeof(struct sockaddr_in));
  vnet_assert(err, "vif_init: bind");
  err = setsockopt(sd, SOL_SOCKET, SO_KEEPALIVE, (char *) &reuse, sizeof(int));
  vnet_assert(err, "vif_init: setsockopt ALIVE");
 
  /* set socket non-blocking.  */
  flags = fcntl(sd, F_GETFL, 0);
  vnet_assert(flags < 0, "vif_init: getfl");
  flags |= O_NDELAY;
  flags = fcntl(sd, F_SETFL, flags);
  vnet_assert(flags < 0, "vif_init: setfl");

  /* listen on socket */
  err = listen(sd, VIF_MAXIF+1);
  vnet_assert(err, "vif_init: listen");

  /* initialize virtual interface table */
  memset((char *) vit, '\0', sizeof(vit_t));
  vit->vit_ifs[VIF_CONIF].vif_ind = rcd;
  vit->vit_ifs[VIF_CONIF].vif_outd = sd;
  vit->vit_maxifd = rcd > sd ? rcd : sd;

  return;
}

/*
 * vif_maxifd(vit)
 *
 * Finds the largest socket descriptor in vit, stores it in
 * vit->vit_maxifd, and returns it.
*/
static int
vif_maxifd(vit_t *vit)
{
  int i, maxifd;
  
  maxifd = vit->vit_ifs[VIF_CONIF].vif_outd;
  for (i = 0; i <= VIF_MAXIF; i++) {
    if (vit->vit_ifs[i].vif_ind > maxifd) {
      maxifd = vit->vit_ifs[i].vif_ind;
    }
  }
        
  if (vit->vit_ifs[VIF_CONIF].vif_ind > maxifd) {
    maxifd = vit->vit_ifs[VIF_CONIF].vif_ind;
  }

  vit->vit_maxifd = maxifd;

  return(maxifd);
}

/*
 * vif_search(vit, vadp)
 *
 * Looks for vadp in vit. * Returns the index where vadp is found,
 * else -1.  Nothing is modified.
*/
static int
vif_search(vit_t *vit, vin_addr_t *vadp)
{
  int i;
  
  for (i = VIF_MINIF; i <= VIF_MAXIF && !vin_eq(&vit->vit_ifs[i].vif_vad, vadp); i++);
  
  if (i <= VIF_MAXIF && vit->vit_ifs[i].vif_status != VIF_OFFLINE) {
    return i;
  } else {
    return -1;
  }
}

/*
 * vif_avail(vit)
 *
 * Returns index to the first unused slot * (i.e. vif is offline) in
 * vit, or -1.  Nothings is modified.
*/
static int
vif_avail(vit_t *vit)
{
  int i;
  
  for (i = VIF_MINIF; i <= VIF_MAXIF && !vif_isoffline(&vit->vit_ifs[i]); i++);

  return(i <= VIF_MAXIF ? i : -1);
}

/*
 * vif_pullup(vif, buf, len)
 *
 * Reads up to len amount of data from vif into buf.
 * buf must be of size at least len.  Assumes non-blocking socket.
 *
 * Behaves the same as recv(3N).
 * If return value is -1 and errno == EWOULDBLOCK,
 * interpret that as pipe is empty.
*/
int
vif_pullup(vif_t *vif, char *buf, int len)
{
  int err;
  fd_set rset;
  struct timeval timer;

  /* Try to pullup the data first */
  err=recv(vif->vif_ind, buf, len, 0);
  
  /* If data not ready, wait awhile */
  if (err == 0) {
    /* link down by seeder */
    errno = EPIPE;
    return(-1);
  }

  if ((err < 0) && (errno == EWOULDBLOCK)) {
    FD_ZERO(&rset);
    FD_SET(vif->vif_ind, &rset);
    timer.tv_sec = 0;
    timer.tv_usec = VIF_PULLUPUSECS;

    err = select(vif->vif_ind+1, &rset, 0, 0, &timer);
    if (err > 0) {
      if (FD_ISSET(vif->vif_ind, &rset)) {
        err=recv(vif->vif_ind, buf, len, 0);
      } else {
        vnet_assert(1, "vif_pullup: select");
      }
    } else if (err == 0) {
      errno = EWOULDBLOCK;
      err=-1;
    }
  }

  return(err);
}

/*
 * vif_pullupn(vif, buf, len)
 *
 * Try up to VIF_PULLUPMAXTRIES times to read exactly len
 * amount of data from vif into buf.  buf must be of
 * size at least len.  Assumes non-blocking socket.
 *
 * On error, returns -1 and errno is set as in recv(3N). 
 * So if after VIF_PULLUPMAXTRIES not the full len amount
 * has been pulled up, returns -1 with errno = EWOULDBLOCK.
 * Otherwise, returns amount of bytes pulled up, which 
 * could be less than the length requested.
*/
int
vif_pullupn(vif_t *vif, char *buf, int len)
{
  int i, err, left, tot;
  char *bp;

  bp = buf;
  left = len;
  i = 0;
  tot = 0;
  do {
    err = vif_pullup(vif, bp, left);
    if (err < 0) {
      if (errno != EWOULDBLOCK) {
        return(err);
      } 
      i++;
    } else {
      bp += err;
      left -= err;
    }
    tot += bp - &buf[0];
  } while (err && left && i < VIF_PULLUPMAXTRIES);

  return (err < 0 ? err : tot);
}
  
/*
 * vif_linkdown(vif, vit)
 *
 * Closes both descriptors of vif, sets them to 0, and marks link
 * down.  Decrements vit->vit_nvif if an incoming descriptor is
 * closed.  vit->vit_maxifd is adjusted to reflect the current max
 * socket descriptor.
*/
void
vif_linkdown(vif_t *vif, vit_t *vit)
{     
  if (vif->vif_ind) {
    close(vif->vif_ind);
    if (FD_ISSET(vif->vif_ind, &vif_input_rset)) {
      /* in case select already returned it as readable */
      vif_input_nready--;
      FD_CLR(vif->vif_ind, &vif_input_rset);
    }
    vif->vif_ind = 0;
    vit->vit_nvif--;
  }

  if (vif->vif_outd) {
    close(vif->vif_outd);
    vif->vif_outd = 0;
  }

  vif->vif_status = VIF_LINKDOWN;
  vif_maxifd(vit);

  return;
}

/*
 * vif_reconfig(self, vit)
 *
 * Called iff a pkt arrives on the socket connected to the seeder.
 *
 * For each vrouter named in the pkt from seeder, the link to that
 * vrouter is either brought down or up.
 *
 * When a link is brought down, the connections on it are closed and
 * its entry in the vit is marked VIF_LINKDOWN, vit_nvif is
 * decremented, and vit_maxifd is updated if necessary.
 *
 * When a link is brought up, first looks for a vif that already has an
 * incoming connection from the other end.  If not found, use the next
 * empty vif in vit and makes the connection to the other end.  Stores
 * the socket descriptor as the outgoing descriptor of the vif.  Also
 * stores the vad of self on the vif vit_maxifd is updated as
 * necessary.
 *
 * After a vif_reconfig, a signal must be sent to the upper layers so
 * that they can clean up their states (e.g. vrp can tear down routes
 * that depend on the torn down vif).  When the upper layers are done
 * with all of their cleaning up, someone is supposed to call
 * vif_offline to finally tear this vif down.
 *
 * Kills process on error.
*/
static void
vif_reconfig(vin_addr_t *self, vit_t *vit)
{
  vif_t *rcif;
  int sd, err, nodelay;
  int i, j, idx;
  u_short metric;
  vrc_t vrc;
  vin_addr_t *vads, *vadp;
  struct sockaddr_in local, remote;
  int flags, reuse;

  rcif = &vit->vit_ifs[VIF_CONIF];
  /* pull up first chunk of header */
  err = vif_pullupn(rcif, (char *) &vrc, sizeof(vrc_t));
  vnet_assert(err <= 0, "vif_reconfig: reconfig header");
  
  /* if version not expected, close connection */
  if (ntohs(vrc.vrc_vers) != VRC_VERSION) {
    close(rcif->vif_ind);
    rcif->vif_ind = 0;
    vif_maxifd(vit);
    vnet_assert(1, "vif_reconfig: unknown VRC version");
  }
  
  /* pull up the rest of header */
  vrc.vrc_nvad = ntohs(vrc.vrc_nvad);
  if (vrc.vrc_nvad > 1) {
    vads = (vin_addr_t *) malloc((vrc.vrc_nvad-1) * sizeof(vrc_t));
    vnet_assert(!vads, "vif_reconfig: malloc");
    err = vif_pullupn(rcif, (char *) vads, (vrc.vrc_nvad-1) * sizeof(vin_addr_t));
    vnet_assert(err <= 0, "vif_reconfig: rest of reconfig header");
  }
  
  /* for each adddress sent: */
  for (i = 0, vadp = &vrc.vrc_vad; i < vrc.vrc_nvad; vadp = &vads[i++]) {
     
    metric = ntohs(vadp->vad_metric);

    if (metric == VIF_LINKDOWN) {

      /* find spot on vif table */
      idx = vif_search(vit, vadp);
      if (idx > 0) {
        /* tearing down links */
        vif_linkdown(&vit->vit_ifs[idx], vit);
      }
      /* else link already down */

    } else {
      
      /* first check if there already exists a connection to the other end */
      idx = vif_search(vit, vadp);
      if (idx < 0) {
        /* new link */
        idx = vif_avail(vit);
        vnet_assert(idx < 0, "vif_reconfig: vif table full");
      }

      /* create socket. */
      sd = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
      vnet_assert(sd < 0, "vif_reconfig: socket");
        
      /* initialize remote address. */
      memset((char *) &remote, 0, sizeof(struct sockaddr_in));
      remote.sin_family = AF_INET;
      memcpy((char *) &remote.sin_addr, (char *) &vadp->vad_ipv4, sizeof(struct in_addr));
      remote.sin_port = vadp->vad_port;
        
      /* connect to server. */
      err = connect(sd, (struct sockaddr *) &remote, sizeof(struct sockaddr_in));
      vnet_assert(err, "vif_reconfig: connect");
        
      /* store address */
      vadp->vad_metric = htons(VIF_DIRECT);
      memcpy((char *) &vit->vit_ifs[idx].vif_vad, (char *) vadp, sizeof(vin_addr_t));
      
      /* store vif */
      vit->vit_ifs[idx].vif_outd = sd;
      vit->vit_ifs[idx].vif_status |= VIF_OUTGOING;
    }
  }

  if (vrc.vrc_nvad > 1) {
    free(vads);
  }
  return;
}

/*
 * vif_newlink(self, vit)
 *
 * A new connection is initiated by another vrouter.  First looks for
 * a vif that already has an outgoing connection to the other end.  If
 * not found, use the next empty vif in vit and makes the connection
 * to the other end.  After a connection is made, stores the socket
 * descriptor as the incoming descriptor of the vif and stores the vad
 * of the other end.  The port number is set to self's, assuming that
 * all vrouters use the same well-known port number.  vit_nvif is
 * incremented and vit_maxifd is updated if necessary.  Returns 0 on
 * sucess, Kills process on error.
*/
static int
vif_newlink(vin_addr_t *self, vit_t *vit)
{
  int idx;
  int len, td, flags;
  struct sockaddr_in client;
  vin_addr_t vad;
  
  /* accept new connection. */
  len = sizeof(struct sockaddr_in);
  td = accept(vit->vit_ifs[VIF_CONIF].vif_outd, (struct sockaddr *) &client, &len);
  if (td < 0) {
    /* See Stevens' UNP 15.6 */
#ifndef EPROTO
#define EPROTO ECONNABORTED
#endif
    if (errno == EWOULDBLOCK || errno == ECONNABORTED
        || errno == EPROTO || errno == EINTR) {
      return(0);
    } else {
      vnet_assert(td < 0, "vif_newlink: accept");
    }
  }

  /* set socket non-blocking.  */
  flags = fcntl(td, F_GETFL, 0);
  vnet_assert(flags < 0, "vif_newlink: getfl");
  flags |= O_NDELAY;
  flags = fcntl(td, F_SETFL, flags);
  vnet_assert(flags < 0, "vif_newlink: setfl");
  
  /* store the vad of the other end in network byte order */
  memcpy((char *) &vad.vad_ipv4, (char *) &client.sin_addr, sizeof(struct in_addr));
  vad.vad_port = self->vad_port;
  vad.vad_metric = htons(VIF_DIRECT);

  /* does a connection to the other end already exist? */
  idx = vif_search(vit, &vad);
  if (idx < 0) {
    
    /* if not */
    idx = vif_avail(vit);
    vnet_assert(idx < 0, "vif_newlink: vif table full");
  
    /* store address */
    memcpy((char *) &vit->vit_ifs[idx].vif_vad, &vad, sizeof(vin_addr_t));
  }

  vit->vit_ifs[idx].vif_ind = td;
  if (td > vit->vit_maxifd) {
    vit->vit_maxifd = td;
  }

  vit->vit_ifs[idx].vif_status |= VIF_INCOMING;

  vit->vit_nvif++;
  
  return (0);
}

/*
 * vif_offline(vif)
 *
 * Marks a VIF_LINKDOWN vif as VIF_OFFLINE.  
 * Must be called by upper layer protocol after a link down
 * event has been handled.
*/
void
vif_offline(vif_t *vif)
{
  if (vif) {
    vnet_assert(vif->vif_status != VIF_LINKDOWN, "vif_offline: vif not down");
    vif->vif_status = VIF_OFFLINE;
  }
  
  return;
}

/*
 * vif_online(vit, up)
 *
 * Returns the number of VIF_LINKUP vifs
 * in vit.  Change the state of those vifs
 * to VIF_ONLINE and for each such vif,
 * put a pointer to it in the up array.
 * The arg up is a pointer to an array of vif pointers.
 * The array must be of minimum size VIF_MAXIF-VIF_MINIF.
 * Must be called by upper layer protocol upon a link up
 * event.
*/
int
vif_online(vit_t *vit, vif_t **up)
{
  int i, nup;
  vif_t *vif;

  nup = 0;
  for (i = VIF_MINIF; i <= VIF_MAXIF; i++) {
    vif = &vit->vit_ifs[i];

    /* if vif is coming up */
    if (vif->vif_status == VIF_LINKUP) {

      /* store the vif in the up array */
      up[nup] = vif;
      nup++;

      /* change state to VIF_ONLINE */
      vif->vif_status = VIF_ONLINE;
    }
  }

  return(nup);
}
      
/*
 * vif_input(self, vit, timer):
 *
 * Waits for packet arrival or periodic timeout.
 * Demuxes and handles the packet.
*/ 
vif_t *
vif_input(vin_addr_t *self, vit_t *vit, struct timeval *timer)
{
  int i, n, err;
  int lnd, rcd, sd;
  vif_t *vif;
  vih_t vih;
  
  lnd = vit->vit_ifs[VIF_CONIF].vif_outd;
  rcd = vit->vit_ifs[VIF_CONIF].vif_ind;

  if (!vif_input_nready) { /* no ready socket */
    
    /* initialize select's variable */
    FD_ZERO(&vif_input_rset);
    
    /* prepare read set */
    FD_SET(lnd, &vif_input_rset);
    if (rcd) {
      FD_SET(rcd, &vif_input_rset);
    }
    for (i = 1, n=vit->vit_nvif; n && i <= VIF_MAXIF; i++) {
      if (sd = vit->vit_ifs[i].vif_ind) {
        FD_SET(sd, &vif_input_rset);
        n--;
      }
    }
    
    /* wait for new pkt, connection, or timeout */      
    vif_input_nready = select(vit->vit_maxifd+1, &vif_input_rset, 0, 0, timer);
    vnet_assert(vif_input_nready < 0, "vif_input: select");
    
    if (!vif_input_nready) {
      /* timeout */
      return((vif_t *) VIF_TIMEOUT);
    }
  }
  
  if (FD_ISSET(lnd, &vif_input_rset)) {
    /* a new link is initiated */
    FD_CLR(lnd, &vif_input_rset);
    vif_input_nready--;
    vif_newlink(self, vit);
    return((vif_t *) VIF_LINKRC);
  }
  
  if (FD_ISSET(rcd, &vif_input_rset)) {
    /* received a reconfiguration pkt */
    FD_CLR(rcd, &vif_input_rset);
    vif_input_nready--;
    
    err = vif_pullupn(&vit->vit_ifs[VIF_CONIF], (char *) &vih, sizeof(vih_t));
    if (err < 0 && errno == EPIPE) {
      close(vit->vit_ifs[VIF_CONIF].vif_ind);
      vit->vit_ifs[VIF_CONIF].vif_ind = 0;
      vif_maxifd(vit);
      return((vif_t *) VIF_TIMEOUT);
    }
    vnet_assert(err < 0, "vif_input: bad seeder pkt");
    
    if (ntohs(vih.vih_vers) == VIF_VERSION) {
      switch (ntohs(vih.vih_prot)) {
      case VIF_PROTOVRC:
        vif_reconfig(self, vit);
        return((vif_t *) VIF_LINKRC);
        break;
      case VIF_PROTOVIP:
        return(&vit->vit_ifs[VIF_CONIF]);
        break;
      }
    }
    vnet_assert(1, "vif_input: bad version or protocol from seeder");
  }
  
  for (i = 0; i <= VIF_MAXIF; i++) {
    /* a pkt arrived on existing link */
    vif = &vit->vit_ifs[i];
    if ((sd = vif->vif_ind) && FD_ISSET(sd, &vif_input_rset)) {
      
      vif_input_nready--;
      FD_CLR(sd, &vif_input_rset);
      
      err = vif_pullupn(vif, (char *) &vih, sizeof(vih_t));
      if (err < 0 && errno == EPIPE) {
        vif_linkdown(vif, vit);
        vif_maxifd(vit);
        return((vif_t *) VIF_LINKRC);
      }
      vnet_assert(err < 0, "vif_input: bad vrouter pkt");
      
      if (ntohs(vih.vih_vers) == VIF_VERSION) {
        if (ntohs(vih.vih_prot) == VIF_PROTOVIP) {
          return(vif);
        }
      }
      fprintf(stderr,
              "vif_input: bad version (%d) or protocol (%d) from vrouter",
              ntohs(vih.vih_vers), ntohs(vih.vih_prot));
      vif_linkdown(vif, vit);
    }
  }

  /* Got a socket that shouldn't be active anymore.
     We're confused.  Reset bits and call select again. */
  FD_ZERO(&vif_input_rset);
  vif_input_nready = 0;
  /* vnet_assert(1, "vif_input: end of function"); */
}

/*
 * vif_output(vif, prot, vmbp);
 *
 * Sends vmbuf out of vif after prepending
 * a vif header with protocol prot.
 * 
 * If no error, returns the amount of data actually sent
 * not including the size of the vif header.  
 * vif not up is considered an error.  On error, behaves like send(3N).
*/
int
vif_output(vif_t *vif, u_short prot, vmbuf_t *vmbp)
{
  int sd;
  int i, plen, left;
  int err;
  extern int errno;
  vih_t vih;
  vmbuf_t *pkt, vmbuf;
  char *buf;

  vnet_assert(!vmbp, "vif_output: empty pkt");
    
  if (!vif_isonline(vif)) {
    errno = EPIPE;
    return(-1);
  }
  
  vih.vih_vers = htons(VIF_VERSION);
  vih.vih_prot = htons(prot);
  
  vmbuf.vmbuf_data = (char *) &vih;
  vmbuf.vmbuf_len = sizeof(vih_t);
  vmbuf.vmbuf_next = vmbp;

  plen = 0;
  for (pkt = &vmbuf; pkt; pkt = pkt->vmbuf_next) {
    buf = pkt->vmbuf_data;
    plen += left = pkt->vmbuf_len;
    do {
      err = send(vif->vif_outd, buf, left, 0);
      if (err < 0 && (errno != EWOULDBLOCK)) {
        err = errno;
        close(vif->vif_outd);
        errno = err;
        vif->vif_outd = 0;
        return(-1);
      } else if (err == 0) {
        errno = EPIPE;
        return(-1);
      } else if (err > 0) {
        left -= err;
        buf += err;
      }
    } while (left > 0);
  }

  return(plen - sizeof(vih_t));
}
