/*
 * Copyright (c) 1997, 1998, 2000, 2004 University of Michigan, Ann Arbor.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of Michigan, Ann Arbor. The name of the University 
 * may not be used to endorse or promote products derived from this 
 * software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Author: Sugih Jamin (jamin@eecs.umich.edu)
*/
#ifndef lint
static char rcsid[] =
"@(#) $Id: vip.c,v 1.15 2000/10/13 17:14:53 sugih Exp $ (UM)";
#endif

#include <stdio.h>
#include <sys/errno.h>
#include <sys/types.h>

#include "vnet.h"
#include "vip.h"
#include "vcmp.h"
#include "vrp.h"
#include "vtp.h"

extern int errno;

extern void _vip_input(vif_t *vif, vrouter_t *vrt);
extern void _vip_forw(vrouter_t *vrt, vip_t *vp, vmbuf_t *vmbp);
extern int _vip_output(vrouter_t *vrt, vin_addr_t *dst, u_short ttl, u_short proto, vmbuf_t *vmbp);
extern void _vip_drop(vrouter_t *vrt, vip_t *vp, char *data, u_short errcode);

/*
 * vip_input(vif, vrt)
 *
 * A VIP pkt arrived on the virtual interface vif.
 * If the pkt carries a bad version number,
 * clears the pipe of any remaining bits.
 * This could remove some good pkts, too bad;
 * this is, after all, a connectionless network.
 *
 * If pkt is bona fide, check whether we are the
 * destination.  If so, figures out its next level
 * protocol and pass it up.
 *
 * If pkt is not for us, pack it up in a vmbuf_t and
 * send it along to its destination by calling vip_forw,
 * passing along the virtual router vrt.
 * 
 * Doesn't return anything.
*/
void
vip_input(vif_t *vif, vrouter_t *vrt)
{
  /* pull up vip header */
  /* if unknown version */
    /* clear interface of bits (see comments below) */
    /* dump vip to screen to help debugging */

  /* else, pull up the rest of the packet */

  /* if pkt addressed to self, strip off vip header and either: 
       - give data portion to vcmp_input() if pkt->vp_proto is VIP_PROTOVCMP,
       - give data portion to vtp_input() if pkt->vp_proto is VIP_PROTOVTP,
       - give data portion to vrp_update() if pkt->vp_proto is VIP_PROTOVRP.*/

  /* else forward pkt */
  
  _vip_input(vif, vrt);
  return;


  /*
   * How to clear the interface of bits in case of error:
   *   Use vif_pullup() until the pipe is empty.
   *
   * > The big question is how much do I pull up and how do I make
   * > sure that I don't start pulling up the header for the next packet.
   *
   * You can't be sure that you're not pulling up the header for the
   * next packet.  What we are doing here is called "Panic-mode recovery" 
   * in the parlance of parser construction.  Here's the definition
   * from the dragon book (Aho, Sethi, and Ullman, _Compilers: Principles,
   * Techniques, and Tools_, p. 164), "On discovering an error, the parser 
   * discards input symbols one at a time until one of a designated set of
   * synchronizing tokens is found.  The synchronizing tokens are usually
   * delimiters, such as semicolon or *end*, whose role in the source
   * program is clear."  In our code, we use an empty pipe as the
   * synchronizing "token"/delimiter.  Put another way, consider a
   * push-down automaton, we recognize an empty pipe as the symbol 
   * denoting end of input that allows us to accept the input and pop 
   * the stack.
   *
   * Empty pipe is detected when vif_pullup() returns -1 with errno set to
   * EWOULDBLOCK or EPIPE.
   */
}
  
/*
 * vip_forw(vrt, vp, vmbp)
 *
 * vrt is the virtual router.
 * vp is the original VIP header of this pkt,
 * in network byte order.
 * vmbp is the vmbuf chain of the remainder of
 * the pkt.
 *
 * Decrements the pkt's hop limit.
 * If hop limit exceeded, drop pkt,
 * else, routes pkt.
 * If route found, forwards pkt
 * to next hop, else drop pkt.
 * Send off VCMP error pkts where appropriate.
*/
void
vip_forw(vrouter_t *vrt, vip_t *vp, vmbuf_t *vmbp)
{
  /* decrement hop limit */
  /* drop pkt if hop limit exceeded */
  /* else, route pkt and send it to the next hop if route found. */

  _vip_forw(vrt, vp, vmbp);
  
  return;
}
    
/*
 * vip_output(vrt, dst, ttl, proto, vmbp)
 *
 * Creates a VIP header with the args.
 * Sends it out to dst along with the vmbuf chain vmbp
 * by calling vif_output.  If ttl is 0, figure out
 * the actual number of hop to dst, otherwise, use the
 * given ttl as the hop limit.
 *
 * Return 0 if no error, otherwise -1 and errno set to:
 * EMSGSIZE: pkt too large,
 * EHOSTUNREACH: dst unreachable,
 * or whatever vif_output sets it to.
*/
int
vip_output(vrouter_t *vrt, vin_addr_t *dst, u_short ttl, u_short proto, vmbuf_t *vmbp)
{
  /* find route to destination */
  /* Make a vip packet: create the header and append data after it.
     This is called marshalling/serialization. */
  /* send packet out */

  return(_vip_output(vrt, dst, ttl, proto, vmbp));
}

/*
 * vip_drop(vrt, vp, data, errcode)
 *
 * Prints to screen the content of the VIP header pointed to by vp.
 * Generates a VCMP error message where appropriate, with errcode 
 * as the VCMP error code defined in vcmp.h.  If errcode is VCMP_ELAST, 
 * don't generate a VCMP error message.
 * For debugging purposes, if data is not null, optionally passes it 
 * on to the higher level protocol it belongs to.  
*/ 
void
vip_drop(vrouter_t *vrt, vip_t *vp, char *data, u_short errcode)
{
  _vip_drop(vrt, vp, data, errcode);
  return;
}
