/*
 * Copyright (c) 1997, 1998, 2000, 2001, 2004 University of Michigan, Ann Arbor.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of Michigan, Ann Arbor. The name of the University 
 * may not be used to endorse or promote products derived from this 
 * software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Author: Sugih Jamin (jamin@eecs.umich.edu)
*/
#ifndef lint
static char rcsid[] =
"@(#) $Id: vrouter.c,v 1.5 2001/09/30 20:14:01 sugih Exp $ (UM)";
#endif

#include <stdio.h>
#include <stdlib.h>
#include <sys/errno.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/resource.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include "vnet.h"
#include "vif.h"
#include "vip.h"
#include "vrp.h"
#include "vtp.h"

#define VROUTER_SEEDER "localhost" /* default seeder host */ 

extern int errno;

void
vrouter_usage(char *progname)
{
  fprintf(stderr, "Usage: %s -p <portnum> -S <seeder host> -P <seeder port> -h <routing heuristics>\n", progname);
  exit(errno);
}
    
/*
 * vrouter_args: Parse args.
 *
 * Return 0 on success or errno on failure.
 * On successful return, "seeder" contains the
 * seeder's host name, "seed_port" contains
 * the seeder's port, and "port" contains the
 * port# to use, in host byte-order.
*/
int
vrouter_args(int argc, char *argv[], u_short *port, char **seeder,
          u_short *seed_port, int *heuristics)
{
  extern char *optarg;
  extern int optind;
  char c;

  errno = 0;

  *port = VROUTER_PORT;
  *seeder = VROUTER_SEEDER;
  *seed_port = SDR_PORT;
  *heuristics = VRP_HEURNONE;
  while ((c = getopt(argc, argv, "p:S:P:h:")) != EOF) {
    switch (c) {
    case 'p':
      *port = (u_short) atoi(optarg);
      if (*port < VNET_MINPORT || *port > VNET_MAXPORT) {
        fprintf(stderr, "%s: -p %d, port number out of range.\n", argv[0], *port);
        exit(1);
      }
      break;
    case 'S':
      *seeder = optarg;
      break;
    case 'P':
      *seed_port = (u_short) atoi(optarg);
      if (*seed_port < VNET_MINPORT || *seed_port > VNET_MAXPORT) {
        fprintf(stderr, "%s: -P %d, port number out of range.\n", argv[0], *seed_port);
        exit(1);
      }
      break;
    case 'h':
      *heuristics = atoi(optarg);
      break;
    default:
      errno = EINVAL;
      break;
    }
  }

  return(errno);
}

/*
 * vrouter_register:
 *
 * Registers with seeder at "seeder.seed_port"
 * First bind the socket to own's local address and port,
 * the connect to seeder.  Seeder will use this
 * connection to seed the virtual network topology.
 *
 * On error, kills process.
 * On success, returns the bound and connected socket descriptor.
*/
int
vrouter_register(vin_addr_t *self, char *seeder, u_short seed_port, u_short port)
{
  int sd, reuse, err;
  struct hostent *sp;
  struct sockaddr_in sdr, local;
  
  /* create a TCP socket */
  sd = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
  vnet_assert(sd < 0, "vrouter_register: socket");
 
  /* initialize own socket address */
  memset((char *) &local, 0, sizeof(struct sockaddr_in));
  local.sin_family = AF_INET;
  memcpy((char *) &local.sin_addr, (char *) &self->vad_ipv4, sizeof(struct in_addr));
  local.sin_port = htons((u_short) port);
 
  /* reuse local address so that bind doesn't complain
     of address already in use. */
  reuse = 1;
  err = setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, (char *) &reuse, sizeof(int));
  vnet_assert(err, "vrouter_register: setsockopt");
  
  /* bind address to socket */
  err = bind(sd, (struct sockaddr *) &local, sizeof(struct sockaddr_in));
  vnet_assert(err, "vrouter_register: bind");
 
  /* initialize seeder's socket address. */
  memset((char *) &sdr, 0, sizeof(struct sockaddr_in));
  sdr.sin_family = AF_INET;
  sdr.sin_port = htons(seed_port);
  sp = gethostbyname(seeder);
  vnet_assert(!sp, "vrouter_register: gethostbyname");
  memcpy(&sdr.sin_addr, sp->h_addr, sp->h_length);

  /* connect to seeder. */
  err = connect(sd, (struct sockaddr *) &sdr, sizeof(struct sockaddr_in));
  vnet_assert(err, "vrouter_register: connect");

  return(sd);
}  

main(int argc, char *argv[])
{
  static int test_count;
  int sd, heuristics;
  u_short port;
  char *seeder;
  u_short seed_port;
  vif_t *vif;
  vrouter_t vrt;
  struct timeval timer;
  vin_addr_t *self;
  struct rlimit coresize;
  extern void vtp_test();
  
  if (vrouter_args(argc, argv, &port, &seeder, &seed_port, &heuristics)) {
    vrouter_usage(argv[0]);
  }
  fprintf(stdout, "%s -p %d -S %s -P %d -h 0x%x\n", argv[0], port, seeder, seed_port, heuristics);

  /* make sure that a core file will be generated
     if this process crashes.  To help in debugging only. */
  getrlimit(RLIMIT_CORE, &coresize);
  coresize.rlim_cur = RLIM_INFINITY;
  setrlimit(RLIMIT_CORE, &coresize);
  
  /* initialize route table. */
  vrp_init(&vrt, port, heuristics);

  self = vrp_self(&vrt);
  
  /* register self with seeder */
  sd = vrouter_register(self, seeder, seed_port, port);
  
  /* initialize virtual interfaces */
  vif_init(sd, port, &vrt.vrouter_vifs);

  srand48((long) self->vad_ipv4.s_addr);
  timer.tv_sec = VRP_UPDTSECS/2;
  timer.tv_usec = 0;
  
  while (1) {
    /* wait for packet */
    vif = vif_input(self, &vrt.vrouter_vifs, &timer);

    switch ((int) vif) {
    case VIF_TIMEOUT:
      break;
    case VIF_LINKRC:
      vrp_reconfig(&vrt);
      break;
    default:
      /* call vip to handle input */
      vip_input(vif, &vrt);
    }
    vrp_periodic(&vrt, &timer);

#ifdef VNET_PA1
#define VNET_MINLOOP 4

    if (test_count++ > VNET_MINLOOP && vrt.vrouter_vifs.vit_nvif == 1) {
      /* if this vrouter has only 1 interface,
       * which in PA1 means it is at the end of the
       * topology chain, have it send some vtp pkts.
       */
      vtp_test(&vrt);
    }
#endif
  }
  
  exit(0);
}
