/*
 * Copyright (c) 1997, 1998, 2000, 2001, 2004 University of Michigan, Ann Arbor.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of Michigan, Ann Arbor. The name of the University 
 * may not be used to endorse or promote products derived from this 
 * software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Author: Sugih Jamin (jamin@eecs.umich.edu)
*/
#ifndef lint
static char rcsid[] =
"@(#) $Id: vrp.c,v 1.18 2001/09/30 20:14:01 sugih Exp sugih $ (UM)";
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include "vnet.h"
#include "vif.h"
#include "vip.h"
#include "vrp.h"

extern void _vrp_update(vrouter_t *vrt, vif_t *vif, vrp_t *vp, int plen);
extern void _vrp_periodic(vrouter_t *vrt, struct timeval *tp);
extern void _vrp_propagate(vrouter_t *vrt);
extern void _vrp_reconfig(vrouter_t *vrt);

/*
 * void
 * vrp_init(vrt, port, heuristics)
 *
 * Takes as argument an uninitialized virtual router and the port number 
 * that makes up the vin_addr_t of this router.  
 * Dump core on error.  Returns nothing.
 * All the fields in the vrouter_t is initialized to the values of
 * this virtual router.  Memory is allocated for the routing table.
*/
void
vrp_init(vrouter_t *vrt, int port, int heuristics)
{
  int err;
  char hname[VRP_HNAMELEN];
  struct hostent *self;
  long local;

  vnet_assert(!vrt, "vrp_init: called with NULL vrt.");
  
  /* implement route table as a simple array */
  vrt->vrouter_rt = (vrtbl_t *) calloc(VRT_INITSIZE, sizeof(vrtbl_t));
  vnet_assert(!vrt->vrouter_rt, "vrp_init: vrtbl calloc");

  /* my host name */
  err = gethostname(hname, VRP_HNAMELEN);
  vnet_assert(err, "vrp_init: gethostname");
  
  /* initialize virtual router.  Initially routing table is empty.
     Assumes the host has only a single IP address. */
  self = gethostbyname(hname);
  vnet_assert(!self, "vrp_init: gethostbyname");
  memcpy((char *) &vrt->vrouter_rt[0].vrtbl_dst.vad_ipv4, 
         (char *) self->h_addr, sizeof(struct in_addr));
  vrt->vrouter_rt[0].vrtbl_dst.vad_port = htons((u_short)port); 
                      /* vin_addr kept in network byte order */

  /* This route is also known as "localhost".  This allows
     for single host testing. */
  local = inet_addr("127.0.0.1");
  memcpy((char *) &vrt->vrouter_local.vad_ipv4, (char *) &local, sizeof(struct in_addr));
  vrt->vrouter_local.vad_port = htons((u_short)port);

  vrt->vrouter_nent = VRT_INITSIZE;
  vrt->vrouter_nmod = 0;
  vrt->vrouter_heuristics = heuristics;

  return;
}

/*
 * void
 * vrp_drop(vrp_t *vp, int plen)
 *
 * This function is for debugging purposes only.  It prints out a vrp
 * pkt that is to be dropped.  vp points to the vrp pkt (in network
 * byte order), plen contains the length of the vrp pkt.
*/
void
vrp_drop(vrp_t *vp, int plen)
{
  int i, ndv, type;
  vin_addr_t *dvp;

  type = ntohs(vp->vrp_type);
  fprintf(stderr, "ver: %d, type: %d\n", ntohs(vp->vrp_ver), type);

  /* print out the pkt anyway, to help debugging */
  switch(type) {
  case VRP_TYPEDV:
    dvp = (vin_addr_t *) &vp->vrp_dv;
    
    ndv = vrp_ndv(plen);
    for (i = 0; i < ndv; i++) {
      fprintf(stderr, "addr: %s\n", inet_ntoa(dvp->vad_ipv4));
      fprintf(stderr, "port: %d; cost: %d\n", ntohs(dvp->vad_port), ntohs(dvp->vad_metric));
      dvp++;
    }
    break;
  case VRP_TYPEPV:
  case VRP_TYPELS:
  default:
    fprintf(stderr, "Type unknown or not currently supported\n");
    break;
  }
  
  return;
}

/*
 * int
 * vrp_rand_dest(vrouter_t *vrt): 
 *
 * Returns a random index into the routing table with valid entry.  
 * Used mainly to randomly pick a destination to send test packets to.  
 * See usage in vtp_test for an example.
*/
int
vrp_rand_dest(vrouter_t *vrt)
{
  int i, j;

  j = 1;
  for (i = 1; i < vrt->vrouter_nent; i++) {
    if (vif_isonline(vrt->vrouter_rt[i].vrtbl_vif)) {
      j = i;
      if (lrand48()%10 > 4) {
        return(i);
      }
    }
  }
  return(j);
}

/*
 * vif_t *
 * vrp_route(vrouter_t *vrt, vin_addr_t *vad, u_short *cost)
 *
 * Search route table vrt for the vin_addr_t in vadp.
 * Return a pointer to the vif_t that goes towards vad
 * If a route is not found, or the cost
 * to destination has metric VRP_INFINITY,
 * return NULL.
 *
 * If "cost" is not a null pointer,
 * store the cost to vad in "cost".
*/
vif_t *
vrp_route(vrouter_t *vrt, vin_addr_t *vad, u_short *cost)
{
  int i;

  /* Look up destination in route table */
  for (i = 1;
       i < vrt->vrouter_nent && 
       (!vrt->vrouter_rt[i].vrtbl_vif || 
        !vin_eq(&vrt->vrouter_rt[i].vrtbl_dst, vad));
       i++);

  /* If found and cost is not VRP_INFINITY, returns the vif.
     Otherwise, returns NULL */
  if (i == vrt->vrouter_nent || vrt->vrouter_rt[i].vrtbl_cost == VRP_INFINITY) {
    if (cost) {
      *cost =  (u_short) VRP_INFINITY;
    }
    return((vif_t *) 0);
  } else {
    if (cost) {
      *cost = (u_short) vrt->vrouter_rt[i].vrtbl_cost;
    }
    return(vrt->vrouter_rt[i].vrtbl_vif);
  }
}

/*
 * void
 * vrp_dump(vrouter_t *vrt):
 *
 * Prints in-use (vrt_cost non zero) route table entries to stderr.
*/
void
vrp_dump(vrouter_t *vrt)
{
  int i;
  struct timeval ts;
  int diff;
  vin_addr_t *vadp;
  char buffer[20];

#ifdef VNET_PA1
  return;
#endif
  
  fprintf(stderr, "VRT for %s.%d (%d avail. entries, last periodic: %d:%d)\n",
          inet_ntoa(vrt->vrouter_rt[0].vrtbl_dst.vad_ipv4), 
          ntohs(vrt->vrouter_rt[0].vrtbl_dst.vad_port),
          vrt->vrouter_nent, vrt->vrouter_ts.tv_sec, vrt->vrouter_ts.tv_usec);
  fprintf(stderr, "Destination          | Next Hop             | Timestamp        | Mod | Cost | Infl | Hold \n");
  fprintf(stderr, "---------------------+----------------------+------------------+-----+------+------+------\n");

  gettimeofday(&ts, NULL);

  for (i = 1; i < vrt->vrouter_nent; i++) {
    if (vrt->vrouter_rt[i].vrtbl_cost) {
      sprintf(buffer, "%s.%d | ",
              inet_ntoa(vrt->vrouter_rt[i].vrtbl_dst.vad_ipv4), 
              ntohs(vrt->vrouter_rt[i].vrtbl_dst.vad_port));
      fprintf(stderr, "%-20s", buffer);

      if (vrt->vrouter_rt[i].vrtbl_vif) {
        vadp = &vrt->vrouter_rt[i].vrtbl_vif->vif_vad;
        sprintf(buffer, "%s.%d | ",
                inet_ntoa(vadp->vad_ipv4), ntohs(vadp->vad_port));
        fprintf(stderr, "%-20s", buffer);
      } else {
        fprintf(stderr, "*link down*          | ");
      } 
      fprintf(stderr, "%9d:%6d |  %d  |  %2d  |  %2d  |  ",
              vrt->vrouter_rt[i].vrtbl_ts.tv_sec, 
              vrt->vrouter_rt[i].vrtbl_ts.tv_usec, 
              vrt->vrouter_rt[i].vrtbl_mod,
              vrt->vrouter_rt[i].vrtbl_cost, 
              vrt->vrouter_rt[i].vrtbl_infl);
      diff = vrt->vrouter_rt[i].vrtbl_hold.tv_sec ? 
             time_diff(vrt->vrouter_rt[i].vrtbl_hold, ts) : 0;
      fprintf(stderr, "%2d\n", diff/VRP_UPDTSECS);
    }
  }
  fprintf(stderr, "\n");

  _vrp_dump(vrt);

  return;    
}

/*
 * void
 * vrp_propagate(vrouter_t *vrt)
 *
 * Called to propagate route updates. vrt is the virtual router.
 * Assumes vrt->vrouter_nmod contains the number of entries to be propagated.
 * Each entry to be propagated must have its vrtbl_mod non-NULL.
 *
 * Allocate space for a VRP pkt of size big enough to hold
 * all the entries to be propagated.  Walks down the route table.
 * For each entry marked for propagation, copy it to the VRP pkt.
 * Sends a copy of the VRP pkt to each neighbor.
 * Implements split horizon with poisonous reverse.
 * Also implements route poisoning if the cost of a destination
 * has been going up for VRP_INFLTRND times.
 * Unmarks all entries marked for propagation.
*/
void
vrp_propagate(vrouter_t *vrt)
{
  /* For each entry marked for propagation, create a distance
     vector in the format of the VRP packet structure. */
    
  /* customize vrp to each neighbor by split horizon with poisonous reverse.
     You know that a node is a neighbor if its cost metric is VIF_DIRECT. */
            
  /* Implement route poisoning if path cost has been going up or */
  /* split horizon with poisonous reverse if sending to next hop */
            
  /* send out vrp to neighbor */

  _vrp_propagate(vrt);
  
  return;
}

 /*
 * void
 * vrp_reconfig(vrt)
 *
 * Called by vrouter when network topology got reconfigured
 * by the seeder.  (In the real world, called when a link
 * goes up or down.) vrt this virtual router.
 *
 * For each vif in the route table, calls vif_isdown to check
 * if the vif has gone down.  If so mark all destinations
 * that uses this vif as the outgoing link (next hop) as
 * having cost VRP_INFINITY.  Initiate path hold down and
 * route poisoning where necessary.  Then for each vif that
 * went down, call vif_offline to let the vif layer knows
 * that the event has been dealt with.
 *
 * Call vif_online to find out the number of links that
 * has come up and to get the vifs.  Add new neighbors
 * to route table.  If neighbor already exists in route
 * table, make it directly connected and fresh, ignoring
 * path hold down.
 * Dump core and exit if route table overflows.
 *
 * If link down and triggered update is selected,
 * call vrp_propagate.
*/ 
void
vrp_reconfig(vrouter_t *vrt)
{
  int i, j;
  vif_t *vif;
  struct timeval ts;
  vif_t *up[VIF_MAXIF];
  int nup;
  vin_addr_t *vadp;

  /* comment the following two lines out once you have
     your implementation of the routing heuristics below
     (where it says PA2). */
  _vrp_reconfig(vrt);
  return;        
  
  /* start of real function, comment out the above two lines
     once you get the routing heuristics implemented 
     where it says PA2 below. */
  gettimeofday(&ts, NULL);
  vrt->vrouter_nmod = 0;

  for (i = 1; i < vrt->vrouter_nent; i++) {

    /* for each valid entry */
    vif = vrt->vrouter_rt[i].vrtbl_vif;
    if (vif) {
      if (vif_isdown(vif)) {
        for (j = 1; j < vrt->vrouter_nent; j++) {
          /* mark all destinations that use
             this vif as unreachable */
          if (vif == vrt->vrouter_rt[j].vrtbl_vif) {
            vrt->vrouter_rt[j].vrtbl_vif = 0;
  
            /* PA2: update link cost, modified time,
                    and implement the necessary
                    routing heuristics. */
          }
        }
        vif_offline(vif);
      }
    }
  }

  /* bring all half-way up line fully up. */
  nup = vif_online(&vrt->vrouter_vifs, up);
  while (nup--) {
    vif = up[nup];
    vadp = &vif->vif_vad;
    
    /* Look up destination in route table */
    for (j = 1; j < vrt->vrouter_nent && 
         !vin_eq(&vrt->vrouter_rt[j].vrtbl_dst, vadp); j++);
    
    if (j == vrt->vrouter_nent) {
      /* not found */
      for (i = 1; i < vrt->vrouter_nent && vrt->vrouter_rt[i].vrtbl_cost; i++);
      vnet_assert(i == vrt->vrouter_nent, 
                  "vrp_reconfig: routing table too small");
      vrt->vrouter_rt[i].vrtbl_vif = vif;
      memcpy((char *) &vrt->vrouter_rt[i].vrtbl_dst, 
             (char *) vadp, sizeof(vin_addr_t));
      vrt->vrouter_rt[i].vrtbl_cost = ntohs(vif->vif_vad.vad_metric);
      vrt->vrouter_rt[i].vrtbl_ts.tv_sec = ts.tv_sec;
      vrt->vrouter_rt[i].vrtbl_ts.tv_usec = ts.tv_usec;
    } else {
      vrt->vrouter_rt[j].vrtbl_vif = vif;
      vrt->vrouter_rt[j].vrtbl_cost = ntohs(vif->vif_vad.vad_metric);
      vrt->vrouter_rt[j].vrtbl_hold.tv_sec = 0;
      vrt->vrouter_rt[j].vrtbl_hold.tv_usec = 0;
      vrt->vrouter_rt[j].vrtbl_infl = 0;
      vrt->vrouter_rt[j].vrtbl_ts.tv_sec = ts.tv_sec;
      vrt->vrouter_rt[j].vrtbl_ts.tv_usec = ts.tv_usec;
    }
  }

  if (vrt->vrouter_nmod) { 
    /* if any entry was modified due to triggered update,
       send it out. */
    vrp_propagate(vrt);
  }

  return;
}
     
/*
 * void
 * vrp_update(vrouter_t *vrt, vif_t *vif, vrp_t *vp, int plen)
 *
 * Called by vip_input when a VRP pkt arrives.
 * vrt is the route table, vif is the vif the pkt arrives on.
 * vp points to the incoming VRP pkt of size plen;
 * the VRP pkt is still in network byte order.
 *
 * For each entry in vp, refreshes the entry, updates its cost, or
 * adds it to route table depending upon whether there is already an
 * entry for the destination in the route table and on the new cost of
 * the destination.  The vad_metric of each vin_addr_t in the arriving
 * VRP pkt is in network byte order.  Translates it to host byte-order
 * before using it for comparison or storing it in the route table.
 * 
 * Path hold down is implemented in this function.
 * If the cost of a destination has been going up, mark it for
 * route poisoning.
 * If cost of a route increases, initiate triggered update by
 * calling vrp_propagate.
 *
 * Dump core and exit if route table overflows.
*/
void
vrp_update(vrouter_t *vrt, vif_t *vif, vrp_t *vp, int plen)
{
  /* drop pkt if vrp_ver or vrp_type is unknown. */

  /* for each dv in the vrp pkt, find dv in routing table */
      /* if direct neighbor (cost is 0 in vad_metric),
         set cost to VIF_DIRECT.  Cost cannot be larger than VRP_INFINITY. */
  /* refresh entry, update cost (initiating path hold down
     and route poisoning when necessary), or add entry */
              /* when there's bouncing effect, the next update 
                 from next hop still contains cost computed from 
                 one's own previous udpate.  So, in implementing 
                 route poisoning we need to discount one update 
                 after seeing the cost went up.  Check vrtbl_reflected
                 for whether one's own previous update has
                 been "reflected" by one's next hop. */
  /* refresh entry only if cost in dv is not infinity */

  /* If cost of any entry increased, do a triggered update */

  _vrp_update(vrt, vif, vp, plen);

  return;
}

/*
 * void
 * vrp_periodic(vrouter_t *vrt, struct timeval *tp)
 *
 * Called periodically by vrouter to send out route updates to neighbors.
 * Returns if it is not yet time to send a periodic update.
 * vrt is the vrouter structure, tp points to the timer whose value
 * determines the next time vrouter will most likely call vrp_periodic again.
 * To avoid synchronized updates, the value of the timer may be
 * changed to VRP_UPDTSECS +- some small random value.
 * 
 * Changes the cost of all entries that have not been refreshed for
 * VRP_STALERND*VRP_UPDTSECS seconds to VRP_INFINITY.
 * Removes all entries that have not been refreshed for
 * (VRP_PURGERND+VRP_STALERND)*VRP_UPDTSECS seconds.
 *
 * Calls vrp_propagate to propagate the periodic updates.
*/
void
vrp_periodic(vrouter_t *vrt, struct timeval *tp)
{
  /* Less than VRP_UPDTSECS has passed between
     last propagate time and current time, do nothing and return */
    
  /* refresh own entry at neighbors */
    /* for each valid entry */
      /* mark each link as needing propagation */
        /* entry has not been refreshed for a long time, stale it */
        /* entry has been stale for a long time, purge it */

  /* Set routing table last propagation time to current time. */
  _vrp_periodic(vrt, tp);

  return;
}

