/*
 * Copyright (c) 1997, 1998, 2000, 2001 University of Michigan, Ann Arbor.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of Michigan, Ann Arbor. The name of the University 
 * may not be used to endorse or promote products derived from this 
 * software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Author: Sugih Jamin (jamin@eecs.umich.edu)
 *
 * @(#) $Id: vrp.h,v 1.18 2001/09/30 20:14:01 sugih Exp sugih $ (UM)
 *
*/
#ifndef __vrp_h__
#define __vrp_h__

#include <sys/time.h>

#include "vnet.h"
#include "vif.h"
#include "vip.h"

typedef struct _vrp vrp_t;       /* vrp pkt header */
typedef struct _vrt_tbl vrtbl_t; /* route table */
typedef struct _vrouter vrouter_t;  /* info on this vrouter */

/* VRP: virtual routing protocol */
#define VRP_VERSION 2        /* VRP version number */
#define VRP_MAXPKTSIZE (VIP_MAXPKTSIZE-sizeof(vrp_t)+sizeof(vin_addr_t))

#define VRT_INITSIZE 11
#define VRT_MAXSIZE  65535
#define VRP_HNAMELEN 80

#define VRP_INFINITY 16      /* definition of infinity */
#define VRP_UPDTSECS 10      /* update period, in seconds */

#define VRP_HEURNONE 0x0     /* No heuristics */
#define VRP_HEURTRGD 0x1     /* Triggered updates */
#define VRP_HEURSHPR 0x2     /* Split Horizon with Poisonous Reverse */
#define VRP_HEURPHD  0x4     /* Path Hold down */
#define VRP_HEURRP   0x8     /* Route Poisoning */
#define VRP_HEURALL  0xf     /* All of the above */

#ifdef VNET_PA1
#define VRP_STALERND 100     /* Set to "infinity" for PA1: */
#else
#define VRP_STALERND 6       /* a route is stale if it has not been
                                refreshed for VRP_STALERND*VRP_UPDTSECS */
#endif
#define VRP_PURGERND 4       /* stale route is kept for
                                VRP_PURGERND*VRP_UPDTSECS longer */
#define VRP_INFLTRND 3       /* do route poisoning if route metric has
                                been increasing for the past VRP_INFLTRND
                                consecutive updates from the same next hop. */
#define VRP_HOLDNRND 2       /* number of updates to hold down path */

#define VRP_TYPEDV 1         /* VRP pkt carries distance vectors */
#define VRP_TYPEPV 2         /* VRP pkt carries path vectors */
#define VRP_TYPELS 3         /* VRP pkt carries link states */

struct _vrp {
  u_short vrp_ver;           /* vrp version */
  u_short vrp_type;          /* VRP_TYPE{DV,PV,LS} */
  vin_addr_t vrp_dv;         /* first of an array of distance vectors,
                                vad metric is cost to destination */
};

struct _vrt_tbl {
  vin_addr_t vrtbl_dst;      /* destination's vin_addr,
                                metric is path cost, in number of hops.
                                vrtbl_dst.vad_metric is ALWAYS kept in
                                host byte order.  This is the only
                                exception from the rule that all fields
                                in vin_addr_t are always in network byte order.
                                Except for vrtbl_dst.vad_metric, the rest of
                                vrtbl_dst is ALWAYS in network byte order*/
  vif_t *vrtbl_vif;          /* interface to next hop.  If vrtbl_vif is NULL,
                                and vrtbl_cost below is 0, that means this entry
                                is uninitialized.  If vrtbl_vif is NULL but
                                vrtbl_cost is VRP_INFINITY, that means this 
                                entry is initialized but the link is 
                                presently down. */
  u_short vrtbl_mod;         /* whether entry has been modified */
  u_short vrtbl_infl;        /* inflation step */
  u_short vrtbl_reflected;   /* when there's bouncing effect, the next
                                update from next hop still contains
                                cost computed from one's own previous
                                udpate.  So, in implementing route
                                poisoning we need to discount one
                                update after seeing the cost went up.
                                This variable is used to keep track of
                                whether one's own previous update has
                                been "reflected" by one's next hop. */
  struct timeval vrtbl_ts;   /* last refresh time of this entry */
  struct timeval vrtbl_hold; /* start of hold down time */
};
#define vrtbl_cost vrtbl_dst.vad_metric
                             /* in host byte order.
                                Except for entry 0, for which cost is always 0,
                                assume a route table entry is currently
                                unused if its vrtbl_cost is zero. */

struct _vrouter {
  vin_addr_t vrouter_local;    /* 127.0.0.1 */

  int vrouter_heuristics;      /* routing algorithm heuristics */
  u_short vrouter_nent;        /* number of entries in routing table */
  u_short vrouter_nmod;        /* number of entries modified */
  struct timeval vrouter_ts;   /* last periodic propagate time */
  vrtbl_t *vrouter_rt;         /* pointer to route table, 
                                  first entry used to keep own's address */
  vit_t vrouter_vifs;          /* pointer to vifs table.  DO NOT write
                                  vrp code to process this table.  Your
                                  vrp code should use ONLY the information
                                  in the vrouter_rt routing table. */
};

/*
 * int
 * vrp_nvif(vrouter_t *vrt)
 *
 * Returns the number of vifs a vrouter has.
 * Modifies nothing.
*/
#define vrp_nvif(vrt) ((vrt) ? vit_nvif(&((vrt)->vrouter_vifs)) : 0)

/*
 * vin_addr_t *
 * vrp_self(vrouter_t *vrt)
 *
 * The vad of this vrouter is assumed to be at 
 * position 0 of the route table.
 * Returns pointer to vad of this vrouter.
 * Nothing is modified.
*/
#define vrp_self(vrt) (&((vrt)->vrouter_rt[0].vrtbl_dst))

/*
 * int
 * vrp_isself(vrouter_t *vrt, vin_addr_t *vad)
 *
 * Returns 1 if the address in vad matches that of
 * this vrouter, 0 otherwise.
 * Nothing is modified.
*/
#define vrp_isself(vrt, vad) \
  (vin_eq(&((vrt)->vrouter_rt[0].vrtbl_dst), vad) || \
   vin_eq(&((vrt)->vrouter_local), vad))

/*
 * vif_t *
 * vrp_route(vrouter_t *vrt, vin_addr_t *vad, u_short *cost)
 *
 * Search route table vrt for the vin_addr_t in vadp.
 * Return a pointer to the vif_t that goes towards vad
 * If a route is not found, or the cost
 * to destination has metric VRP_INFINITY,
 * return NULL.
 *
 * If "cost" is not a null pointer,
 * store the cost to vad in "cost".
*/
extern vif_t *vrp_route(vrouter_t *vrt, vin_addr_t *vad, u_short *cost);

/*
 * void
 * vrp_update(vrouter_t *vrt, vif_t *vif, vrp_t *vp, int plen)
 *
 * vrt is the vrouter.
 * vif is the vif the update pkt came in from.
 * vp points to the VRP packet.
 * plen is the length of the VRP pkt vp points to.
*/
extern void vrp_update(vrouter_t *vrt, vif_t *vif, vrp_t *vp, int plen);

/*
 * int
 * vrp_rand_dest(vrouter_t *vrt): 
 *
 * Returns a random index into the routing table with valid entry.  
 * Used mainly to randomly pick a destination to send test packets to.  
 * See usage in vtp_test for an example.
*/
extern int vrp_rand_dest(vrouter_t *vrt);

/*
 * void
 * vrp_drop(vrp_t *vp, int plen)
 *
 * This function is for debugging purposes only.
 * It prints out a vrp pkt that is to be dropped.
 * vp points to the vrp pkt (in network byte order),
 * plen contains the length of the vrp pkt.
*/
extern void vrp_drop(vrp_t *vp, int plen);

/*
 * int
 * vrp_ndv(int len)
 *
 * Given the length of a VRP packet, returns the number of distance
 * vector the packet carries.
*/
#define vrp_ndv(len) ((len - sizeof(vrp_t) + sizeof(vin_addr_t))/sizeof(vin_addr_t))

extern void vrp_init(vrouter_t *vrt, int port, int heuristics);
extern void vrp_periodic(vrouter_t *vrt, struct timeval *tp);
extern void vrp_reconfig(vrouter_t *vrt);
extern void vrp_dump(vrouter_t *vrt);

#endif /*__vrp_h__*/
